import csv, os, argparse

import rdkit
from rdkit import DataStructs, Chem
from rdkit.Chem import AllChem
from rdkit.SimDivFilters.rdSimDivPickers import MaxMinPicker

try:
    from rdkit.Chem import inchi as rdinchi
    inchi_check = True
except Exception as e:
    inchi_check = False
    print("RDkit inchi class not found.\n To use the -tautomer you have to enable InChi support in RDkit (http://www.rdkit.org/docs/Install.html).\n")

try:
    import standardiser
    from standardiser import standardise as standardise_rdmol
    standardiser_check = True
except Exception as e:
    standardiser_check = False
    print("standardiser library not found. You can get it at https://github.com/flatkinson/standardiser.\n")

class Reader():
    """
    Class takes a path to a file and according to its extension tries to process it.

    Parameters
    ----------
    filepath : str
        Path to an input file.
    standardise : bool
        When set to False only canonization of smiles representation is used to uniform a molecule.
        When set to True standardiser library (https://github.com/flatkinson/standardiser) is used to process a molecule (neutralise, unsalt...).
    remove_tautomers: bool
        Use InChi representation of molecules to remove tautomers.
    """

    def __init__(self, filepath, standardise, remove_tautomers):
        self.filepath = filepath
        self.standardise = standardise
        self.remove_tautomers = remove_tautomers
        self.repair = True
        self.ext = self.filepath.split(".")[-1]
        self.ext2fnc = {
            "csv": self._read_csv_file,
            "txt": self._read_csv_file,
            "smi": self._read_csv_file,
            "sdf": self._read_sdf_file,
        }

        if not self.ext in self.ext2fnc:
            return

        else:
            self.ext2fnc[self.ext](self.filepath)

        if self.standardise:
            if standardiser_check:
                self._standardise_structures()
            else:
                raise Exception("standardiser library not found.\n To use the -standardise please get the standardiser library at https://github.com/flatkinson/standardiser.\n")

        if self.remove_tautomers:
            if inchi_check:
                self._remove_tautomers()
            else:
                raise Exception("RDkit inchi class not found.\n To use the -tautomer you have to enable InChi support in RDkit (http://www.rdkit.org/docs/Install.html).\n")

        print("Distinct molecules: {}".format(len(set(self.cansmi2rdmol.keys()))))
        print("\n")


    def _read_csv_file(self, filepath):
        print("Reading file: {}".format(filepath))
        with open(filepath, "r") as r:
            if self.ext == "smi":
                self.data = [row for row in csv.reader(r)]
                self.fields = {"smiles_index": 0, "id_index": False, "header": False}
            else:
                dialect = csv.Sniffer().sniff(r.read(), delimiters=";,")
                r.seek(0)
                self.data = [row for row in csv.reader(r, delimiter=dialect.delimiter, quotechar=dialect.quotechar, lineterminator=dialect.lineterminator)]
                
                if len(self.data) > 1:
                    first_rows = self.data[:5] if len(self.data) > 4 else self.data[:2]
                    self.fields = self._guess_fields(first_rows)

        self._read_molecules()

    def _remove_tautomers(self):
        print("Removing tautomers...")
        self.cansmi2inchi = {}
        inchi2cansmi = {}
        cansmi_set = set()

        for molid, cansmi in self.id2cansmi.items():
            if not cansmi in cansmi_set:
                cansmi_set.add(cansmi)

                inchi = rdinchi.MolToInchi(self.cansmi2rdmol[cansmi])
                self.cansmi2inchi[cansmi] = inchi

                if inchi in inchi2cansmi:
                    self.id2cansmi[molid] = inchi2cansmi[inchi]
                    del self.cansmi2rdmol[cansmi]
                else:
                    inchi2cansmi[inchi] = cansmi

    def _standardise_structures(self):
        print("Standardising structures...")
        cansmi2rdmol = {}

        for key, cansmi in self.id2cansmi.items():
            try:
                std_rdmol = standardise_rdmol.apply(self.cansmi2rdmol[cansmi])
                std_cansmi = Chem.MolToSmiles(std_rdmol)
                
                if std_cansmi != cansmi:
                    self.id2cansmi[key] = std_cansmi
                    cansmi2rdmol[std_cansmi] = std_rdmol
                else:
                    cansmi2rdmol[cansmi] = self.cansmi2rdmol[cansmi]

            except Exception as e:
                cansmi2rdmol[cansmi] = self.cansmi2rdmol[cansmi]

        self.cansmi2rdmol = cansmi2rdmol


    def _get_cansmi2fp(self):
        """Returns dict with canonical smiles as keys and RDkit Circular Morgan fingerprint object with radius 2"""
        self.cansmi2fp = {cansmi: AllChem.GetMorganFingerprint(rdmol, radius=2) for cansmi, rdmol in self.cansmi2rdmol.items()}
        return self.cansmi2fp


    def _read_molecules(self):
        print("Reading molecules...")
        rows = self.data[1:] if self.fields["header"] else self.data
        self.cansmi2rdmol = {}
        self.id2cansmi = {}
        self.defective = []

        for i, r in enumerate(rows, 1):
            # molid = r[self.fields["id_index"]] if self.fields["id_index"] is not False else i
            molid = i
            mol = Chem.MolFromSmiles(r[self.fields["smiles_index"]])

            if mol is not None:
                cansmi = Chem.MolToSmiles(mol)
                self.id2cansmi[molid] = cansmi
                self.cansmi2rdmol[cansmi] = mol
            else:
                if self.repair:
                    smi = self._repair_smiles(r[self.fields["smiles_index"]])
                    if smi != r[self.fields["smiles_index"]]:
                        mol = Chem.MolFromSmiles(smi)

                        if mol:
                            cansmi = Chem.MolToSmiles(mol)
                            self.id2cansmi[molid] = cansmi
                            self.cansmi2rdmol[cansmi] = mol
                
                if not mol:
                    self.defective.append(molid)

        print("Input rows: {}".format(len(rows)))
        print("Molecules: {}".format(len(self.id2cansmi)))
        print("Not parsed: {}".format(len(self.defective)))

    def _repair_smiles(self, smi):
        chars = [",", ";"]
        
        for char in chars:
            if char in smi:
                smi = ".".join(smi.split(char))

        smi = "".join(smi.split(" "))
        return smi

    def _read_sdf_file(self, filepath):
        print("Reading file: {}".format(filepath))
        self.cansmi2rdmol = {}
        self.id2cansmi = {}
        self.defective = []

        mols = [m for m in Chem.SDMolSupplier(filepath)]

        for i, mol in enumerate(mols, 1):
            if mol:
                cansmi = Chem.MolToSmiles(mol)
                self.cansmi2rdmol[cansmi] = mol
                self.id2cansmi[i] = cansmi
            else:
                self.defective.append(i)

        print("Input molecules: {}".format(len(mols)))
        print("Molecules: {}".format(len(self.id2cansmi)))
        print("Not parsed: {}".format(len(self.defective)))

    def _guess_fields(self, rows):
        smiles_fields = ["smiles", "smi"]
        id_fields = ["id", "name"]
        smiles_index = False
        id_index = False
        header = False

        for i, item in enumerate(rows[0]):
            if item.lower() in smiles_fields:
                smiles_index = i
                header = True

            if item.lower() in id_fields:
                id_index = i
                header = True

        if smiles_index is False:
            smiles_index = self._guess_smiles_index(rows[:1])

            if smiles_index is not False:
                header = False

            else:
                header = True
                smiles_index = self._guess_smiles_index(rows)

        return {"id_index": id_index, "smiles_index": smiles_index, "header": header}

    def _guess_smiles_index(self, rows):
        smiles_index = False
        indexes = []
        for row in rows:
            for i, item in enumerate(row):
                if item != "":
                    try:
                        mol = Chem.MolFromSmiles(item)
                        if mol:
                            indexes.append(i)
                    except Exception as e:
                        pass
        
        if len(set(indexes)) == 1:
            smiles_index = indexes[0]

        elif len(set(indexes)) > 1:
            smiles_index = indexes[0]
            count = 0
            
            for i in set(indexes):
                i_count = indexes.count(i)
                if i_count > count:
                    count = i_count
                    smiles_index = i

        return smiles_index


class MolSetInspector():
	"""
	MolSetInspector class facilitates the processing of multiple molecular sets in different text file formats.
	It takes a directory containing molecular sets in files (.sdf, .csv, .smi, .txt) which are read and joined in one set of distinct molecules.
	The molecules are canonicalized, can be further standardised (neutralised, unsalted etc.) and tautomers removed.
	As a result MolSetInspector can write the molecular sets interesections, wrong (not parsed) molecules IDs and distinct molecules 
	along with the hit table (hit table shows in which set/s was the molecule found). 
	It can also filter the distinct molecules by their diversity in two ways: 
	by the maximum total number of diverse molecules or by the maximum similarity of molecular pair in the set.
	"""

	def __init__(self, indirectory, outdirectory=False, standardise=False, remove_tautomers=False):
		"""
		Parameters
		----------
		indirectory : str
		   Input directory containing molecular sets (files).
		outdirectory : str
			Output directory used for writing files.
		standardise: bool
			Use standardiser (https://github.com/flatkinson/standardiser) when processing molecules.
		remove_tautomers: bool
			Use InChi representation of molecules to remove tautomers.
		"""
		self.indirectory = indirectory
		self.outdirectory = outdirectory
		self.standardise = standardise
		self.remove_tautomers = remove_tautomers

		self.files = os.listdir(self.indirectory)
		self.file2reader = {}
		
		if len(self.files) == 0:
			print("There are no files in directory: {}".format(self.indirectory))
		else:
			for f in self.files:
				filepath = os.path.join(self.indirectory, f)
				self.file2reader[f] = Reader(filepath, self.standardise, self.remove_tautomers)

	def get_set_intersections(self):
		"""
		Returns matrix in the form of array of arrays containing intersection counts of all pairs of molecular sets.
		"""
		print("Set intersections...")
		header = [""]
		header.extend(self.files)

		data = [header]

		for f1 in self.files:
			row = [f1]
			s1 = set(self.file2reader[f1].id2cansmi.values())
			for f2 in self.files:
				s2 = set(self.file2reader[f2].id2cansmi.values())
				row.append(len(s1 & s2))

			data.append(row)

		if self.outdirectory:
			self._write_file(data, "intersections.csv")

		return data

	def get_defective(self):
		"""
		Returns a list containing set name and position (starting at 1) af a defective molecule in a set.
		"""
		header = ["set", "position"]
		data = [header]

		for f in self.files:
			for m in self.file2reader[f].defective:
				data.append([f, m])

		print("Defective molecules: {}".format(len(data)-1))

		if self.outdirectory:
			self._write_file(data, "defective.csv")

		return data

	def get_distinct(self, diversity_type=False, treshold=False, output_format="csv", hit_table=False):
		"""
		Returns a list of distinct molecules from all processed sets. 
		It enables to filter molecules by diversity using maximum similarity treshold or total number of molecules.

		Parameters
		----------
		diversity_type : str
			Diversity by maximum similarity (max_similarity) or by total number (total_number) of diverse molecules.
		treshold : number
			Treshold for diversity selection. From 0 to 1 for max_similarity and positive integer for total_number.
		output_format : str
			Output format of the distinct set (sdf, smi, csv).
		hit_table : bool
			Only when csv output_format is used.
			Append information about the presence of molecule in each set.
		"""
		print("Distinct molecules...")

		distinct = set()

		if self.remove_tautomers:
			inchis = set()
			
			for f, r in self.file2reader.items():
				for cansmi, inchi in r.cansmi2inchi.items():
					if inchi not in inchis and cansmi in r.cansmi2rdmol:
						inchis.add(inchi)
						distinct.add(cansmi)
		else:
			for f, r in self.file2reader.items():
				distinct = distinct | set(r.id2cansmi.values())


		print("Distinct molecules: {}".format(len(distinct)))
		
		if diversity_type != False and treshold != False:

			if diversity_type == "max_similarity":
				distinct = self._get_diverse_by_maximum_similarity(distinct, treshold)
			elif diversity_type == "total_number" and treshold < len(distinct):
				distinct = self._get_diverse_by_total_number(distinct, treshold)
		
		data = [[d] for d in distinct]

		if output_format == "csv":
			header = ["smiles"]
			
			if hit_table:
				for f in self.files:
					header.append(f)
					id2cansmi = self.file2reader[f].id2cansmi
					intersection = distinct & set(id2cansmi.values())

					for i, d in enumerate(data):
						if d[0] in intersection:
							data[i].append(1)
						else:
							data[i].append("")

				header.append("sum")
				
				for i, d in enumerate(data):
					data[i].append(d.count(1))

			result = [header]
			result.extend(data)

			if self.outdirectory:
				self._write_file(result, "distinct.csv")

		elif output_format == "smi":
			result = data

			if self.outdirectory:
				self._write_file(result, "distinct.smi")

		elif output_format == "sdf":
			result = []

			for f in self.files:
				for cansmi, rdmol in self.file2reader[f].cansmi2rdmol.items():
					if cansmi in distinct:
						result.append(Chem.MolToMolBlock(rdmol))

			if self.outdirectory:
				self._write_sdf_file(result, "distinct.sdf")

		return result

	def _get_cansmi2fp(self, distinct_cansmi):
		cansmi2fp = {}

		for f, r in self.file2reader.items():
			cansmi2fp.update(r._get_cansmi2fp())

		return cansmi2fp

	def _write_file(self, data, filename):
		print("Writing file: {}".format(filename))
		self._checkdir(self.outdirectory)
		with open(os.path.join(self.outdirectory, filename), "w") as w:
			writer = csv.writer(w)
			writer.writerows(data)

	def _write_sdf_file(self, molblock_list, filename):
		print("Writing SDF file: {}".format(filename))
		self._checkdir(self.outdirectory)
		with open(os.path.join(self.outdirectory, filename), "w") as w:
			w.write("\n$$$$\n".join(molblock_list))

	def _get_diverse_by_total_number(self, distinct_cansmi, treshold):
		print("Diverse molecules by total number: {}".format(treshold))
		cansmi2fp = self._get_cansmi2fp(distinct_cansmi)
		ordered_cansmi = [c for c in distinct_cansmi]
		fps = [cansmi2fp[c] for c in ordered_cansmi]

		def distance(fp1, fp2, fps=fps):
			return 1-DataStructs.TanimotoSimilarity(fps[fp1], fps[fp2])
		
		picker = MaxMinPicker()
		pick_indexes = picker.LazyPick(distance, len(ordered_cansmi), treshold)
		distinct = set([ordered_cansmi[i] for i in pick_indexes])
		return distinct

	def _get_diverse_by_maximum_similarity(self, distinct_cansmi, treshold):
		print("Diverse molecules by maximum similarity: {}".format(treshold))
		cansmi2fp = self._get_cansmi2fp(distinct_cansmi)
		ordered_cansmi = [c for c in distinct_cansmi]
		fps = [cansmi2fp[c] for c in ordered_cansmi]
		distinct = set()
		count = len(ordered_cansmi)
		
		for i, c1 in enumerate(ordered_cansmi):
			fp1 = cansmi2fp[c1]
			add_cansmi = True

			for j in xrange(i+1, count):
				fp2 = cansmi2fp[ordered_cansmi[j]]
				sim = self._tanimoto_similarity(fp1, fp2)
				
				if sim > treshold:
					add_cansmi = False
					break

			if add_cansmi:
				distinct.add(c1)
		
		print("Diverse molecules: {}".format(len(distinct)))
		return distinct

	def _tanimoto_similarity(self, fp1, fp2):
		return DataStructs.TanimotoSimilarity(fp1, fp2)

	def _checkdir(self, directory):
		if directory and not (os.path.isfile(directory) or os.path.exists(directory)):
			os.makedirs(directory)

def _process_(arguments):
    s = MolSetInspector(arguments.indirectory, arguments.output_directory, arguments.standardise_structures, arguments.remove_tautomers)

    if arguments.distinct:
    	if arguments.diverse_by_max_similarity or arguments.diverse_by_total_number:
    		if arguments.diverse_by_max_similarity and arguments.diverse_by_total_number:
    			raise Exception("Only diverse_by_max_similarity or diverse_by_total_number can be used. Not both at the same time.")

    		if arguments.diverse_by_max_similarity:
    			s.get_distinct(diversity_type="max_similarity", treshold=arguments.diverse_by_max_similarity, output_format=arguments.output_format, hit_table=arguments.hit_table)
    		else:
    			s.get_distinct(diversity_type="total_number", treshold=arguments.diverse_by_total_number, output_format=arguments.output_format, hit_table=arguments.hit_table)
    	else:
    		s.get_distinct(output_format=arguments.output_format, hit_table=arguments.hit_table)
	
	if arguments.defective_molecules:
		s.get_defective()

    if arguments.intersections:
    	s.get_set_intersections()

if __name__ == '__main__':
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument("indirectory", type=str, help="directory containing molecular sets files")    
    parser.add_argument("-o", "--output_directory", default=False, type=str, help="output directory")    
    parser.add_argument("-outf", "--output_format", type=str, help="output format for distinct molecules (smi, sdf, csv)", default="csv")    
    parser.add_argument("-dist", "--distinct", default=False, help="write file with distinct molecules from all sets", action="store_true")
    parser.add_argument("-hit", "--hit_table", default=False, help="append table with 0/1 to distinct export which shows the sets that contains the molecule", action="store_true")
    parser.add_argument("-inter", "--intersections", default=False, help="write file with set intersections", action="store_true")
    parser.add_argument("-defect", "--defective_molecules", default=False, help="write file with molecules that couldn't be parsed", action="store_true")
    parser.add_argument("-standardise", "--standardise_structures", default=False, help="standardise structures using the standardiser", action="store_true")
    parser.add_argument("-tautomer", "--remove_tautomers", default=False, help="remove tautomers", action="store_true")
    parser.add_argument("-dbs", "--diverse_by_max_similarity", type=float, default=False, help="get diverse structures by maximum similarity treshold (from 0 to 1, 1 is most similar)")
    parser.add_argument("-dbn", "--diverse_by_total_number", type=int, default=False, help="get maximum specified number of diverse structures (positive integer)")
    
    args = parser.parse_args()
    _process_(args)
